import asyncio
from typing import TYPE_CHECKING
from http.cookies import SimpleCookie
from base64 import b64encode

from aiohttp import ClientResponseError
from rsa import PublicKey, encrypt

from .exceptions import LoginError, ApiError
from .constants import STEAM_URL
from .utils import get_cookie_value_from_session, generate_session_id

if TYPE_CHECKING:
    from .client import SteamCommunityMixin

__all__ = ("LoginMixin", "REFERER_HEADER")

REFERER_HEADER = {"Referer": str(STEAM_URL.COMMUNITY) + "/"}
API_HEADERS = {
    "accept": "application/json, text/plain, */*",
    "sec-fetch-site": "cross-site",
    "sec-fetch-mode": "cors",
    "sec-fetch-dest": "empty",
}


class LoginMixin:
    """Mixin with login logic methods."""

    __slots__ = ()

    _is_logged: bool
    _refresh_token: str | None
    # Almost useless, but ...
    # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L230
    _access_token: str | None

    def __init__(self, *args, refresh_token: str = None, access_token: str = None, **kwargs):
        self._is_logged = False
        self._refresh_token = refresh_token
        self._access_token = access_token

        super().__init__(*args, **kwargs)

    @property
    def is_logged(self) -> bool:
        return self._is_logged

    @property
    def session_id(self: "SteamCommunityMixin") -> str | None:
        return get_cookie_value_from_session(self.session, STEAM_URL.COMMUNITY, "sessionid")

    async def is_session_alive(self: "SteamCommunityMixin") -> bool:
        r = await self.session.get(STEAM_URL.COMMUNITY)
        rt = await r.text()
        return self.username.lower() in rt.lower()

    async def login(self: "SteamCommunityMixin", *, init_data=True, init_session=True):
        """
        Perform login.
        Populate `api_key`, `trade_token`, `wallet_country`, `wallet_currency` fields if it is required.

        :param init_data: fetch initial required data (api key, trade token, wallet_currency/wallet_country)
        :param init_session: init session before start auth process.
            Set this to False if you already make requests to steam from current client.
        :raises ApiError: when failed to obtain rsa key, update steam guard code
        :raises LoginError: other login process errors
        """

        # https://github.com/bukson/steampy/blob/fe0433c8cf7020318cfbbc22e79028a7576374ee/steampy/login.py#L67
        # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/examples/login-to-web-with-2fa.ts#L13
        init_session and await self.session.get(STEAM_URL.COMMUNITY)

        session_data = await self._begin_auth_session_with_credentials()
        await self._update_auth_session_with_steam_guard_code(session_data)
        await self._poll_auth_session_status(session_data)
        fin_data = await self._finalize_login()  # there can be retrieved steam id

        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/LoginSession.ts#L845
        loop = asyncio.get_event_loop()
        transfers = [
            loop.create_task(self._perform_transfer(d, fin_data["steamID"])) for d in fin_data["transfer_info"]
        ]
        done, _ = await asyncio.wait(transfers, return_when=asyncio.FIRST_COMPLETED)

        # Transfers exception check ?

        self._set_web_cookies(done.pop().result())

        init_data and await self._init_data()

        self._is_logged = True

    async def _perform_transfer(self: "SteamCommunityMixin", data: dict, steam_id: str | int = None) -> SimpleCookie:
        """
        Perform a transfer of params and tokens to steam login endpoints.
        Similar behavior to arrow function from link below.

        .. seealso:: https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L845-L868
        """

        r = await self.session.post(data["url"], data={**data["params"], "steamID": steam_id or self.steam_id})
        # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L864
        # make sure that `steamLoginSecure` cookie is present
        if not r.cookies.get("steamLoginSecure"):
            raise ApiError("No `steamLoginSecure` cookie in result.")

        return r.cookies

    def _set_web_cookies(self: "SteamCommunityMixin", cookie: SimpleCookie):
        """
        Set web cookies to session for main steam urls.

        .. seealso:: https://github.com/DoctorMcKay/node-steamcommunity/blob/7c564c1453a5ac413d9312b8cf8fe86e7578b309/index.js#L153-L175
        """

        # ensure that sessionid cookie is presented
        # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L872-L873
        if not cookie.get("sessionid"):
            cookie["sessionid"] = generate_session_id()

        for k, morsel in cookie.items():
            for url in (STEAM_URL.STORE, STEAM_URL.COMMUNITY, STEAM_URL.HELP):
                c = SimpleCookie()
                m = morsel.copy()
                m["domain"] = url.host
                c[k] = m
                self.session.cookie_jar.update_cookies(c, response_url=url)

    async def _begin_auth_session_with_credentials(self: "SteamCommunityMixin") -> dict:
        pub_key, ts = await self._get_rsa_key()
        # for web browser
        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/AuthenticationClient.ts#L390
        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/enums-steam/EAuthTokenPlatformType.ts
        platform_data = {
            "website_id": "Community",
            "device_details": {
                "device_friendly_name": self.user_agent,
                "platform_type": 2,
            },
        }

        data = {
            "account_name": self.username,
            "encrypted_password": b64encode(encrypt(self._password.encode("utf-8"), pub_key)).decode(),
            "encryption_timestamp": ts,
            "remember_login": "true",
            "persistence": "1",
            **platform_data,
        }
        r = await self.session.post(
            STEAM_URL.API.IAuthService.BeginAuthSessionViaCredentials,
            data=data,
            headers=REFERER_HEADER,
        )
        return await r.json()

    async def _update_auth_session_with_steam_guard_code(self: "SteamCommunityMixin", session_data: dict):
        # Doesn't check allowed confirmations, but it's probably not needed
        # as steam accounts suited for trading must have a steam guard and device code.

        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/LoginSession.ts#L735
        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/enums-steam/EAuthSessionGuardType.ts
        data = {
            "client_id": session_data["response"]["client_id"],
            "steamid": session_data["response"]["steamid"],
            "code_type": 3,
            "code": self.two_factor_code,
        }

        try:
            await self.session.post(
                STEAM_URL.API.IAuthService.UpdateAuthSessionWithSteamGuardCode,
                data=data,
                headers=REFERER_HEADER,
            )
        except ClientResponseError:
            raise ApiError("Error updating steam guard code")

    async def _poll_auth_session_status(self: "SteamCommunityMixin", session_resp: dict):
        data = {
            "client_id": session_resp["response"]["client_id"],
            "request_id": session_resp["response"]["request_id"],
        }
        r = await self.session.post(
            STEAM_URL.API.IAuthService.PollAuthSessionStatus,
            data=data,
            headers=REFERER_HEADER,
        )
        rj = await r.json()
        if rj.get("response", {"had_remote_interaction": True})["had_remote_interaction"]:
            raise ApiError("Error polling auth session status.", rj)

        self._refresh_token = rj["response"]["refresh_token"]
        self._access_token = rj["response"]["access_token"]

    async def _finalize_login(self: "SteamCommunityMixin") -> dict:
        data = {
            "nonce": self._refresh_token,
            "sessionid": self.session_id,
            "redir": str(STEAM_URL.COMMUNITY / "login/home/?goto="),
        }
        r = await self.session.post(
            STEAM_URL.LOGIN / "jwt/finalizelogin",
            data=data,
            headers={**API_HEADERS, **REFERER_HEADER},
        )
        rj: dict = await r.json()

        if rj and rj.get("error"):
            raise LoginError("Get error response when performing login finalization.", rj)
        elif not rj or not rj.get("transfer_info"):
            raise LoginError("Malformed login response.", rj)

        return rj

    async def _get_rsa_key(self: "SteamCommunityMixin") -> tuple[PublicKey, int]:
        r = await self.session.get(STEAM_URL.API.IAuthService.GetPasswordRSAPublicKey % {"account_name": self.username})
        rj = await r.json()
        try:
            rsa_mod = int(rj["response"]["publickey_mod"], 16)
            rsa_exp = int(rj["response"]["publickey_exp"], 16)
            rsa_timestamp = int(rj["response"]["timestamp"])

            return PublicKey(rsa_mod, rsa_exp), rsa_timestamp

        except KeyError:
            raise ApiError("Could not obtain rsa-key.", rj)

    async def logout(self: "SteamCommunityMixin") -> None:
        await self.session.post(
            STEAM_URL.COMMUNITY / "login/logout/",
            data={**REFERER_HEADER, "sessionid": self.session_id},
        )
        self._is_logged = False

    # async def refresh_access_token(self: "SteamCommunityMixin") -> str:
    #     """"""
    #
    #     # TODO this
    #     r = await self.session.post(
    #         STEAM_URL.API.IAuthService.GenerateAccessTokenForApp,
    #         data={"refresh_token": self._refresh_token, "steamid": self.steam_id},
    #         headers={**API_HEADERS, **REFERER_HEADER},
    #     )
    #     rj = await r.json()
    #
    #     try:
    #         self._access_token = rj["response"]["access_token"]
    #     except KeyError:
    #         raise ApiError("Can't renew access token.", rj)
    #
    #     return self._access_token



"""Useful utils."""

import asyncio
from base64 import b64decode, b64encode
from struct import pack, unpack
from time import time as time_time
from hmac import new as hmac_new
from hashlib import sha1
from functools import wraps
from typing import Callable, overload, ParamSpec, TypeVar, TYPE_CHECKING, TypeAlias
from http.cookies import SimpleCookie, Morsel
from math import floor
from secrets import token_hex
from re import search as re_search, compile as re_compile
from json import loads as j_loads

from aiohttp import ClientSession, ClientResponse
from yarl import URL

from .typed import JWTToken

if TYPE_CHECKING:
    from .client import SteamCommunityMixin

__all__ = (
    "gen_two_factor_code",
    "generate_confirmation_key",
    "generate_device_id",
    "extract_openid_payload",
    "do_session_steam_auth",
    "get_cookie_value_from_session",
    "async_throttle",
    "create_ident_code",
    "account_id_to_steam_id",
    "steam_id_to_account_id",
    "id64_to_id32",
    "id32_to_id64",
    "to_int_boolean",
    "restore_from_cookies",
    "get_jsonable_cookies",
    "update_session_cookies",
    "buyer_pays_to_receive",
    "receive_to_buyer_pays",
    "generate_session_id",
    "decode_jwt",
    "find_item_nameid_in_text",
)


def gen_two_factor_code(shared_secret: str, timestamp: int = None) -> str:
    """Generate twofactor (onetime/TOTP) code."""

    if timestamp is None:
        timestamp = int(time_time())
    time_buffer = pack(">Q", timestamp // 30)  # pack as Big endian, uint64
    time_hmac = hmac_new(b64decode(shared_secret), time_buffer, digestmod=sha1).digest()
    begin = ord(time_hmac[19:20]) & 0xF
    full_code = unpack(">I", time_hmac[begin : begin + 4])[0] & 0x7FFFFFFF  # unpack as Big endian uint32
    chars = "23456789BCDFGHJKMNPQRTVWXY"
    code = ""

    for _ in range(5):
        full_code, i = divmod(full_code, len(chars))
        code += chars[i]

    return code


def generate_confirmation_key(identity_secret: str, tag: str, timestamp: int = None) -> str:
    if timestamp is None:
        timestamp = int(time_time())
    buff = pack(">Q", timestamp) + tag.encode("ascii")
    return b64encode(hmac_new(b64decode(identity_secret), buff, digestmod=sha1).digest()).decode()


# It works, however it's different that one generated from mobile app
def generate_device_id(steam_id: int) -> str:
    """Generate mobile android device id. Confirmation endpoints requires this."""

    hexed_steam_id = sha1(str(steam_id).encode("ascii")).hexdigest()
    return "android:" + "-".join(
        [hexed_steam_id[:8], hexed_steam_id[8:12], hexed_steam_id[12:16], hexed_steam_id[16:20], hexed_steam_id[20:32]]
    )


def extract_openid_payload(page_text: str) -> dict[str, str]:
    """
    Extract steam openid payload (specs) from page html raw text.
    Use it if 3rd party websites have extra or non-cookie auth (JWT via service API call, for ex.).

    :param page_text:
    :return: dict with payload data
    """

    # not so beautiful as with bs4 but dependency free
    return {
        "action": re_search(r"id=\"actionInput\"[\w=\"\s]+value=\"(?P<action>\w+)\"", page_text)["action"],
        "openid.mode": re_search(r"name=\"openid\.mode\"[\w=\"\s]+value=\"(?P<mode>\w+)\"", page_text)["mode"],
        "openidparams": re_search(r"name=\"openidparams\"[\w=\"\s]+value=\"(?P<params>[\w=/]+)\"", page_text)["params"],
        "nonce": re_search(r"name=\"nonce\"[\w=\"\s]+value=\"(?P<nonce>\w+)\"", page_text)["nonce"],
    }


async def do_session_steam_auth(session: ClientSession, auth_url: str | URL) -> ClientResponse:
    """
    Request auth page, find specs of steam openid and log in through steam with passed session.
    Use it when you need to log in 3rd party site trough Steam using only cookies.

    .. seealso:: https://aiosteampy.somespecial.one/examples/auth_3rd_party_site/

    :param session: just session.
    :param auth_url: url to site, which redirect you to steam login page.
    :return: response with history, headers and data
    """

    r = await session.get(auth_url)
    rt = await r.text()

    data = extract_openid_payload(rt)

    return await session.post("https://steamcommunity.com/openid/login", data=data, allow_redirects=True)


def get_cookie_value_from_session(session: ClientSession, url: URL | str, field: str) -> str | None:
    """
    Get value from session cookies.
    Passed `url` must include scheme (for ex. `https://url.com`)
    """

    c = session.cookie_jar.filter_cookies(URL(url))
    return c[field].value if field in c else None


_P = ParamSpec("_P")
_R = TypeVar("_R")


@overload
def async_throttle(seconds: float, *, arg_index: int):
    ...


@overload
def async_throttle(seconds: float, *, arg_name: str):
    ...


@overload
def async_throttle(seconds: float):
    ...


def async_throttle(
    seconds: float,
    *,
    arg_index: int = None,
    arg_name: str = None,
) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    """
    Prevents the decorated function from being called more than once per `seconds`.
    Throttle (`await asyncio.sleep`) before call wrapped async func,
    when related arg was equal (same hash value) to related arg passed in previous func call
    if time between previous and current call lower than `seconds`.
    Related arg must be hashable (can be dict key).
    Wrapped func must be async (return `Coroutine`).
    Throttle every func call time if related arg has been not specified.

    :param seconds: seconds during which call frequency has been limited.
    :param arg_index: index of related arg in *args tuple.
    :param arg_name: keyname of related arg in **kwargs.
    """

    # mega optimization, prevent arg checks in wrapped func call
    # I know about PEP8: E731, but this way is much shorter and readable
    if arg_index is None and arg_name is None:
        get_key = lambda _, __: None
    elif arg_index:
        get_key = lambda a, _: a[arg_index]
    else:
        get_key = lambda _, k: k[arg_name]

    ts_map: dict[..., float] = {}

    def decorator(f: Callable[_P, _R]) -> Callable[_P, _R]:
        @wraps(f)
        async def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            key = get_key(args, kwargs)
            ts = time_time()
            diff = ts - ts_map.get(key, 0)
            if diff < seconds:
                await asyncio.sleep(diff + 0.01)  # +10ms to ensure that second passed since last call
                ts_map[key] = time_time()
            else:
                ts_map[key] = ts

            return await f(*args, **kwargs)

        return wrapper

    return decorator


def create_ident_code(obj_id: int | str, app_id: int | str, context_id: int | str = None) -> str:
    """
    Create unique ident code for :class:`aiosteampy.models.EconItem` asset or item class
    (description) within whole Steam Economy.

    .. seealso:: https://dev.doctormckay.com/topic/332-identifying-steam-items/

    :param obj_id: asset or class id of Steam Economy Item
    :param app_id: app id of Steam Game
    :param context_id: context id of Steam Game. Only for `EconItem`
    :return: ident code
    """

    code = f"{obj_id}_{app_id}"
    if context_id is not None:
        code += f"_{context_id}"

    return code


def steam_id_to_account_id(steam_id: int) -> int:
    """Convert steam id64 to steam id32."""

    return steam_id & 0xFFFFFFFF


def account_id_to_steam_id(account_id: int) -> int:
    """Convert steam id32 to steam id64."""

    return 1 << 56 | 1 << 52 | 1 << 32 | account_id


id64_to_id32 = steam_id_to_account_id
id32_to_id64 = account_id_to_steam_id


def to_int_boolean(s):
    """Convert something to 1, 0."""

    return 1 if s else 0


JSONABLE_COOKIE_JAR: TypeAlias = list[dict[str, dict[str, str, None, bool]]]


def update_session_cookies(cookies: JSONABLE_COOKIE_JAR, session: ClientSession):
    """Update the session cookies from jsonable cookie jar."""

    for cookie_data in cookies:
        c = SimpleCookie()
        for k, v in cookie_data.items():
            copied = dict(**v)  # copy to avoid modification of the arg
            m = Morsel()
            m._value = copied.pop("value")
            m._key = copied.pop("key")
            m._coded_value = copied.pop("coded_value")
            m.update(copied)
            c[k] = m

        session.cookie_jar.update_cookies(c)


async def restore_from_cookies(
    cookies: JSONABLE_COOKIE_JAR,
    client: "SteamCommunityMixin",
    *,
    init_data=True,
    **init_kwargs,
) -> bool:
    """
    Helper func. Restore client session from cookies.
    Login if session is not alive.
    Return `True` if cookies are valid and not expired.
    """

    update_session_cookies(cookies, client.session)

    # find access token
    for cookie_data in cookies:
        for k, v in cookie_data.items():
            if v["key"] == "steamLoginSecure":
                try:
                    client._access_token = v["value"].split("%7C%7C")[1]
                    break
                except IndexError:
                    pass

    if not (await client.is_session_alive()):
        await client.login(init_data=init_data, **init_kwargs)
        return False
    else:
        client._is_logged = True
        init_data and await client._init_data()
        return True


def get_jsonable_cookies(session: ClientSession) -> JSONABLE_COOKIE_JAR:
    """Extract and convert cookies to dict object."""

    return [
        {
            field_key: {
                "coded_value": morsel.coded_value,
                "key": morsel.key,
                "value": morsel.value,
                "expires": morsel["expires"],
                "path": morsel["path"],
                "comment": morsel["comment"],
                "domain": morsel["domain"],
                "max-age": morsel["max-age"],
                "secure": morsel["secure"],
                "httponly": morsel["httponly"],
                "version": morsel["version"],
                "samesite": morsel["samesite"],
            }
            for field_key, morsel in cookie.items()
        }
        for cookie in session.cookie_jar._cookies.values()
    ]


def receive_to_buyer_pays(
    amount: int,
    *,
    publisher_fee=0.1,
    steam_fee=0.05,
    wallet_fee_min=1,
    wallet_fee_base=0,
) -> tuple[int, int, int]:
    """
    Convert `to_receive` amount in `buyer_pays`. Mostly needed for placing sell listing.
    Works just like on sell listing window on steam.

    :param amount: amount in cents
    :return: steam fee value, publisher fee value, buyer pays amount
    """

    steam_fee_value = int(floor(max(amount * steam_fee, wallet_fee_min) + wallet_fee_base))
    publisher_fee_value = int(floor(max(amount * publisher_fee, 1) if publisher_fee > 0 else 0))
    return steam_fee_value, publisher_fee_value, int(amount + steam_fee_value + publisher_fee_value)


def buyer_pays_to_receive(
    amount: int,
    *,
    publisher_fee=0.1,
    steam_fee=0.05,
    wallet_fee_min=1,
    wallet_fee_base=0,
) -> tuple[int, int, int]:
    """
    Convert `buyer_pays` amount in `to_receive`. Mostly needed for placing sell listing.
    Works just like on sell listing window on steam.

    :param amount: amount in cents
    :return: steam fee value, publisher fee value, amount to receive
    """

    # I don't know how it works, it's just a copy of js function working on inputs in steam front
    estimated_amount = int((amount - wallet_fee_base) / (steam_fee + publisher_fee + 1))
    s_fee, p_fee, v = receive_to_buyer_pays(
        estimated_amount,
        publisher_fee=publisher_fee,
        steam_fee=steam_fee,
        wallet_fee_min=wallet_fee_min,
        wallet_fee_base=wallet_fee_base,
    )

    i = 0
    some_flag = False
    while (v != amount) and (i < 10):
        if v > amount:
            if some_flag:
                s_fee, p_fee, v = receive_to_buyer_pays(
                    estimated_amount - 1,
                    publisher_fee=publisher_fee,
                    steam_fee=steam_fee,
                    wallet_fee_min=wallet_fee_min,
                    wallet_fee_base=wallet_fee_base,
                )
                s_fee += amount - v
                v = amount
                break
            else:
                estimated_amount -= 1
        else:
            some_flag = True
            estimated_amount += 1

        s_fee, p_fee, v = receive_to_buyer_pays(
            estimated_amount,
            publisher_fee=publisher_fee,
            steam_fee=steam_fee,
            wallet_fee_min=wallet_fee_min,
            wallet_fee_base=wallet_fee_base,
        )
        i += 1

    return s_fee, p_fee, int(v - s_fee - p_fee)


def generate_session_id() -> str:
    """
    Generate steam like session id.

    .. seealso:: https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L801C19-L801C50
    """

    # Hope ChatGPT knows what she is doing
    return token_hex(12)


def decode_jwt(token: str) -> JWTToken:
    parts = token.split(".")
    if len(parts) != 3:
        raise ValueError("Invalid JWT", parts)

    return j_loads(b64decode(parts[1] + "==", altchars="-_"))


_ITEM_NAMEID_RE = re_compile(r"Market_LoadOrderSpread\(\s?(?P<nameid>\d+)\s?\)")


def find_item_nameid_in_text(text: str) -> int | None:
    """Find and return`item_nameid` in HTML text response from Steam Community Market page"""

    res = _ITEM_NAMEID_RE.search(text)
    return int(res["nameid"]) if res is not None else res



"""
Trade and interact with steam market, webapi, guard.
"""

from .exceptions import ApiError, LoginError, ConfirmationError, SessionExpired
from .constants import Game, STEAM_URL, Currency, Language, TradeOfferStatus, MarketListingStatus
from .client import SteamClient, SteamPublicClient
from .models import (
    MarketListing,
    EconItem,
    Confirmation,
    MarketListingItem,
    MyMarketListing,
    BuyOrder,
    MarketHistoryListing,
    MarketHistoryListingItem,
    MarketHistoryEvent,
    TradeOfferItem,
    TradeOffer,
    HistoryTradeOffer,
    HistoryTradeOfferItem,
)



from re import compile, search as re_search
from urllib.parse import quote
from json import loads
from http.cookies import SimpleCookie

from aiohttp import ClientSession
from aiohttp.client import _RequestContextManager
from yarl import URL

from .http import SteamHTTPTransportMixin
from .guard import SteamGuardMixin
from .confirmation import ConfirmationMixin
from .login import LoginMixin
from .trade import TradeMixin
from .market import MarketMixin
from .public import SteamPublicMixin, INV_PAGE_SIZE, PREDICATE, PRIVATE_USER_EXC_MSG

from .models import Notifications, EconItem
from .typed import WalletInfo
from .constants import STEAM_URL, Currency, GameType, Language
from .exceptions import ApiError, SessionExpired
from .utils import get_cookie_value_from_session, steam_id_to_account_id, account_id_to_steam_id

__all__ = ("SteamClient", "SteamPublicClient", "SteamCommunityMixin")

API_KEY_RE = compile(r"<p>Key: (?P<api_key>[0-9A-F]+)</p>")

API_KEY_CHECK_STR = "<h2>Access Denied</h2>"
API_KEY_CHECK_STR1 = "You must have a validated email address to create a Steam Web API key"
STEAM_LANG_COOKIE = "Steam_Language"
DEF_COUNTRY = "UA"


class SteamCommunityMixin(
    SteamGuardMixin,
    ConfirmationMixin,
    LoginMixin,
    MarketMixin,
    TradeMixin,
    SteamPublicMixin,
    SteamHTTPTransportMixin,
):
    """
    Mixin class, but with `__init__`. Inherits all other mixins.
    Needed in case you want to use multiple inheritance with `__slots__`.
    """

    __slots__ = ()

    username: str
    steam_id: int | None
    trade_token: str

    def __init__(
        self,
        username: str,
        password: str,
        # It is possible to get steam id from the cookie and then arg will not be necessary,
        # but typical use case of the library means that the user already knows the steam id
        steam_id: int,
        *,
        shared_secret: str,
        identity_secret: str,
        refresh_token: str = None,
        access_token: str = None,
        api_key: str = None,
        trade_token: str = None,
        wallet_currency: Currency = None,
        wallet_country=DEF_COUNTRY,
        lang=Language.ENGLISH,
        tz_offset=0.0,
        session: ClientSession = None,
        proxy: str = None,
        user_agent: str = None,
    ):
        self.username = username
        self.steam_id = account_id_to_steam_id(steam_id) if steam_id < 4294967296 else steam_id  # steam id64
        self.trade_token = trade_token

        self._password = password
        self._shared_secret = shared_secret
        self._identity_secret = identity_secret
        self._api_key = api_key
        self._wallet_currency = wallet_currency

        self._wallet_country = wallet_country

        super().__init__(
            session=session,
            user_agent=user_agent,
            access_token=access_token,
            refresh_token=refresh_token,
            proxy=proxy,
        )

        self._set_init_cookies(lang, tz_offset)

    @property
    def account_id(self) -> int:
        """Steam id32."""
        return steam_id_to_account_id(self.steam_id)

    @property
    def trade_url(self) -> URL:
        return STEAM_URL.TRADE / "new/" % {"partner": self.account_id, "token": self.trade_token or ""}

    @property
    def profile_url(self) -> URL:
        return STEAM_URL.COMMUNITY / f"profiles/{self.steam_id}"

    @property
    def language(self) -> Language:
        """Language of Steam html pages, json info, descriptions, etc."""
        return Language(get_cookie_value_from_session(self.session, STEAM_URL.STORE, STEAM_LANG_COOKIE))

    @property
    def country(self) -> str:
        """Just wallet country. Needed for public methods."""
        return self._wallet_country

    @property
    def currency(self) -> Currency:
        """Alias for wallet currency. Needed for public methods."""
        return self._wallet_currency

    def _set_init_cookies(self, lang: str, tz_offset: float):
        urls = (STEAM_URL.COMMUNITY, STEAM_URL.STORE, STEAM_URL.HELP)
        for url in urls:
            c = SimpleCookie()
            c[STEAM_LANG_COOKIE] = lang
            c[STEAM_LANG_COOKIE]["path"] = "/"
            c[STEAM_LANG_COOKIE]["domain"] = url.host
            c[STEAM_LANG_COOKIE]["secure"] = True

            self.session.cookie_jar.update_cookies(cookies=c, response_url=url)

        # Is it needed?
        c_key = "timezoneOffset"
        for url in urls:
            c = SimpleCookie()
            c[c_key] = str(tz_offset).replace(".", ",")
            c[c_key]["path"] = "/"
            c[c_key]["domain"] = url.host
            c[c_key]["SameSite"] = True

            self.session.cookie_jar.update_cookies(cookies=c, response_url=url)

    async def _init_data(self):
        if not self._api_key:
            self._api_key = await self._fetch_api_key()
            not self._api_key and await self.register_new_api_key()

        if not self.trade_token:
            self.trade_token = await self._fetch_trade_token()
            not self.trade_token and await self.register_new_trade_url()

        if not self._wallet_currency:
            wallet_info = await self.fetch_wallet_info()
            self._wallet_country = wallet_info["wallet_country"]
            if not self._wallet_currency:
                self._wallet_currency = Currency(wallet_info["wallet_currency"])

    async def fetch_wallet_info(self) -> WalletInfo:
        """
        Get wallet info from inventory page.

        .. warning:: May reset new items notifications count.

        :return: wallet info
        """

        r = await self.session.get(self.profile_url / "inventory", headers={"Referer": str(self.profile_url)})
        rt = await r.text()
        info: dict = loads(re_search(r"g_rgWalletInfo = (?P<info>.+);", rt)["info"])
        if not info.get("success"):
            raise ApiError("Failed to fetch wallet info from inventory.", info)

        return info

    async def get_wallet_balance(self) -> int:
        """Fetch wallet balance and currency."""

        # Why is this endpoint do not work sometimes?
        r = await self.session.get(STEAM_URL.STORE / "api/getfundwalletinfo")
        rj = await r.json()
        if not rj.get("success"):
            raise ApiError("Failed to fetch wallet info.", rj)

        self._wallet_currency = Currency.by_name(rj["user_wallet"]["currency"])
        return int(rj["user_wallet"]["amount"])

    async def register_new_trade_url(self) -> URL:
        """Register new trade url. Cache token."""

        r = await self.session.post(
            self.profile_url / "tradeoffers/newtradeurl",
            data={"sessionid": self.session_id},
        )
        token: str = await r.json()

        self.trade_token = quote(token, safe="~()*!.'")  # https://stackoverflow.com/a/72449666/19419998
        return self.trade_url

    async def _fetch_trade_token(self) -> str | None:
        r = await self.session.get(self.profile_url / "tradeoffers/privacy")
        rt = await r.text()

        search = re_search(r"\d+&token=(?P<token>.+)\" readonly", rt)
        return search["token"] if search else None

    async def _fetch_api_key(self) -> str | None:
        r = await self.session.get(STEAM_URL.COMMUNITY / "dev/apikey")
        rt = await r.text()
        if API_KEY_CHECK_STR in rt or API_KEY_CHECK_STR1 in rt:
            raise ApiError(API_KEY_CHECK_STR1)

        search = API_KEY_RE.search(rt)
        return search["api_key"] if search else None

    async def register_new_api_key(self, domain="https://github.com/somespecialone/aiosteampy") -> str:
        """
        Register new api key, cache it and return.

        :param domain: On which domain api key will be registered
        :return: api key
        """

        data = {
            "domain": domain,
            "agreeToTerms": "agreed",
            "sessionid": self.session_id,
            "Submit": "Register",
        }
        r = await self.session.post(STEAM_URL.COMMUNITY / "dev/registerkey", data=data)
        rt = await r.text()

        self._api_key = API_KEY_RE.search(rt)["api_key"]
        return self._api_key

    async def get_notifications(self) -> Notifications:
        """Get notifications count."""

        headers = {"Referer": str(self.profile_url)}
        r = await self.session.get(STEAM_URL.COMMUNITY / "actions/GetNotificationCounts", headers=headers)
        rj = await r.json()
        return Notifications(*(rj["notifications"][str(i)] for i in range(1, 12) if i != 7))

    def reset_items_notifications(self) -> _RequestContextManager:
        """
        Fetches self inventory page, which resets new items notifications to 0.

        .. seealso:: https://github.com/DoctorMcKay/node-steamcommunity/blob/7c564c1453a5ac413d9312b8cf8fe86e7578b309/index.js#L275
        """

        return self.session.get(self.profile_url / "inventory/")

    async def get_inventory(
        self,
        game: GameType,
        *,
        predicate: PREDICATE = None,
        page_size=INV_PAGE_SIZE,
    ) -> list[EconItem]:
        """
        Fetches self inventory.

        :param game: just Steam Game
        :param page_size: max items on page. Current Steam limit is 2000
        :param predicate: callable with single arg `EconItem`, must return bool
        :return: list of `EconItem`
        :raises ApiError: for ordinary reasons
        :raises SessionExpired:
        """

        try:
            inv = await self.get_user_inventory(self.steam_id, game, predicate=predicate, page_size=page_size)
        except ApiError as e:
            raise SessionExpired if e.msg == PRIVATE_USER_EXC_MSG else e  # self inventory can't be private
        return inv


class SteamClient(SteamCommunityMixin):
    """Ready to use client class with all inherited methods."""

    __slots__ = (
        "_is_logged",
        "_refresh_token",
        "_access_token",
        "session",
        "username",
        "steam_id",
        "_password",
        "_shared_secret",
        "_identity_secret",
        "_api_key",
        "trade_token",
        "device_id",
        "_wallet_currency",
        "_wallet_country",
    )


class SteamPublicClient(SteamPublicMixin, SteamHTTPTransportMixin):
    """Class for public methods that do not require login."""

    __slots__ = ("session", "language", "currency", "country")

    def __init__(
        self,
        *,
        language=Language.ENGLISH,
        currency=Currency.USD,
        country=DEF_COUNTRY,
        session: ClientSession = None,
        proxy: str = None,
        user_agent: str = None,
    ):
        self.language = language
        self.currency = currency
        self.country = country

        super().__init__(session=session, user_agent=user_agent, proxy=proxy)



"""User agents service extension."""

from collections import UserList
from random import choice

from yarl import URL
from aiohttp import ClientSession

API_URL = URL("https://randua.somespecial.one")


class UserAgentsService(UserList[str]):
    """
    List-like class of user agents responsible for loading and getting random user agents.

    .. seealso:: https://github.com/somespecialone/random-user-agent
    """

    __slots__ = ("_api_url",)

    def __init__(self, *, api_url=API_URL):
        """
        :param api_url: url of `random user agent` backend service api
        """

        super().__init__()

        self._api_url = api_url

    @property
    def agents(self) -> list[str]:
        return self.data

    async def load(self):
        async with ClientSession(raise_for_status=True) as sess:
            r = await sess.get(self._api_url / "all")
            agents: list[str] = (await r.text()).splitlines()

        self.data = agents

    def get_random(self) -> str:
        return choice(self.agents)



from functools import partial

from yarl import URL
from aiohttp import ClientSession, InvalidURL

try:
    from aiohttp_socks import ProxyConnector
except ImportError:
    ProxyConnector = None


__all__ = ("SteamHTTPTransportMixin",)


class SteamHTTPTransportMixin:
    """
    Handler of session instance, helper getters/setters.
    Probably future setup for `web api caller` or something similar

    .. seealso:: https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/transports/WebApiTransport.ts
    """

    __slots__ = ()

    session: ClientSession

    def __init__(self, *args, session: ClientSession = None, proxy: str = None, user_agent: str = None, **kwargs):
        if proxy and session:
            raise ValueError("You need to handle proxy connection by yourself with predefined session instance.")
        elif proxy:
            if "socks" in proxy:
                if ProxyConnector is None:
                    raise TypeError(
                        """
                        To use `socks` type proxies you need `aiohttp_socks` package. 
                        You can do this with `aiosteampy[socks]` dependency install target.
                        """
                    )

                self.session = ClientSession(connector=ProxyConnector.from_url(proxy), raise_for_status=True)
            else:  # http/s
                self.session = ClientSession(raise_for_status=True)

                try:
                    proxy = URL(proxy)
                except ValueError as e:
                    raise InvalidURL(proxy) from e

                self.session._request = partial(self.session._request, proxy=proxy)  # patch session instance

        elif session:
            self.session = session
        else:
            self.session = ClientSession(raise_for_status=True)

        if user_agent:
            self.user_agent = user_agent

        super().__init__(*args, **kwargs)

    @property
    def user_agent(self) -> str | None:
        return self.session.headers.get("User-Agent")

    @user_agent.setter
    def user_agent(self, value: str):
        self.session.headers["User-Agent"] = value
